#import "ONSLController.h"
#import "ONSLGameManager.h"
#import "ONSLGameInfo.h"
#import "DragImageView.h"
#import "ONSLFontManager.h"
#import "ONSLUpdateManager.h"

#define LS(str) NSLocalizedString(str,@"")
#define TableRowType @"row"
#define TableRowTypes [NSArray arrayWithObjects:@"row",nil]

@implementation ONSLController

- (id)init
{
	[super init];
	mgr = [[ONSLGameManager alloc] initWithDelegate:self];
	fontMgr = [[ONSLFontManager alloc] init];
	updateMgr = [[ONSLUpdateManager alloc]initWithDelegate:self];
	updateURL = nil;
	quitOK = NO;
	lastLogStr = nil;
	lastLogTime = 0;
	bufferLine = 0;
	logArr = [[NSMutableArray alloc] init];
	NSFontManager *fm = [NSFontManager sharedFontManager];
	[fm setDelegate:self];
	toolbarItems = [[NSMutableDictionary alloc] init];
	
	return self;
}

- (void)dealloc
{
	[mgr release];
	[fontMgr release];
	[updateMgr release];
	if(updateURL) [updateURL release];
	if(lastLogStr) [lastLogStr release];
	[logArr release];
	[toolbarItems release];
	[super dealloc];
}

- (IBAction)addGame:(id)sender
{
	[o_prefGameIcon setImage:[NSImage imageNamed:@"NSApplicationIcon"]];
	[o_prefGameIcon clearPath];
	[o_prefGameName setStringValue:@"New Game"];
	[o_prefGameDirectory setStringValue:@""];
	[o_prefGameOption setStringValue:@""];
	[o_fullscreenMode setState:NSOffState];
	[o_varEditMode setState:NSOffState];
	[o_wheelAdvanceMode setState:NSOffState];
	[o_prefGameFontName setStringValue:[[NSFont fontWithName:@"Osaka-Mono" size:13] displayName]];
	[o_prefGameFontName setFont:[NSFont fontWithName:@"Osaka-Mono" size:13]];
	[[NSFontPanel sharedFontPanel] setPanelFont:[NSFont fontWithName:@"Osaka-Mono" size:13] isMultiple:NO];
	[NSApp beginSheet:o_gameEditPane
		modalForWindow:o_mainWindow
		modalDelegate:self
		didEndSelector:@selector(sheetDidEnd:returnCode:contextInfo:)
		contextInfo:@"AddGame"];
}

- (IBAction)editGame:(id)sender
{
	ONSLGameInfo *gi = [mgr gameInfoAtIndex:[o_gameListTable selectedRow]];
	[o_prefGameName setStringValue:[gi title]];
	[o_prefGameDirectory setStringValue:[gi path]];
	[o_prefGameOption setStringValue:[gi otherArgs]];
	if([gi fullscreenMode]) [o_fullscreenMode setState:NSOnState];
	else [o_fullscreenMode setState:NSOffState];
	if([gi varEditMode]) [o_varEditMode setState:NSOnState];
	else [o_varEditMode setState:NSOffState];
	if([gi wheelAdvanceMode]) [o_wheelAdvanceMode setState:NSOnState];
	else [o_wheelAdvanceMode setState:NSOffState];
	[o_prefGameFontName setStringValue:[[gi font] displayName]];
	[o_prefGameFontName setFont:[gi font]];
	[[NSFontPanel sharedFontPanel] setPanelFont:[gi font] isMultiple:NO];
	
	if([[gi imagePath] isEqualToString:@""]) {
		[o_prefGameIcon setImage:[NSImage imageNamed:@"NSApplicationIcon"]];
		[o_prefGameIcon clearPath];
	}
	else {
		[o_prefGameIcon setImgPath:[gi imagePath]];
		if([[NSFileManager defaultManager] isReadableFileAtPath:[gi imagePath]])
			[o_prefGameIcon setImage:[[[NSImage alloc] initByReferencingFile:[gi imagePath]] autorelease]];
		else [o_prefGameIcon setImage:[NSImage imageNamed:@"NSApplicationIcon"]];
	}
	[NSApp beginSheet:o_gameEditPane
		modalForWindow:o_mainWindow
		modalDelegate:self
		didEndSelector:@selector(sheetDidEnd:returnCode:contextInfo:)
		contextInfo:@"EditGame"];
	
}

- (IBAction)endEdit:(id)sender
{
	if([sender tag] == 0) {
		if([[o_prefGameName stringValue] isEqualToString:@""]) {
			NSRunCriticalAlertPanel(LS(@"error"), LS(@"name is empty"), LS(@"Cancel"), nil, nil);
			return;
		}
		if([[o_prefGameDirectory stringValue] isEqualToString:@""]) {
			NSRunCriticalAlertPanel(LS(@"error"), LS(@"path is empty"), LS(@"Cancel"), nil, nil);
			return;
		}
	}
	[NSApp endSheet:o_gameEditPane returnCode:[sender tag]];
	[o_gameEditPane makeFirstResponder:o_prefGameName];
	[o_gameEditPane close];
}

- (IBAction)removeGame:(id)sender
{
	NSBeginAlertSheet(LS(@"remove"), @"OK", nil, @"Cancel", o_mainWindow, self, @selector(sheetDidEnd:returnCode:contextInfo:), nil, @"RemoveGame", [NSString stringWithFormat:LS(@"remove ok?"),[[mgr gameInfoAtIndex:[o_gameListTable selectedRow]] title]]);
}

- (IBAction)launchGame:(id)sender
{
	id game = [mgr gameInfoAtIndex:[o_gameListTable selectedRow]];
	if([game isPlaying]) {
		NSBeginCriticalAlertSheet(LS(@"quit"), @"OK", nil, @"Cancel", o_mainWindow, self, @selector(sheetDidEnd:returnCode:contextInfo:), nil, @"Terminate", [NSString stringWithFormat:LS(@"%@: terminate ok?"), [game title]]);
		return;
	}
	else [game execute];
	[self updateTable];
}

- (IBAction)chooseGameDirectory:(id)sender
{
	NSOpenPanel *op = [NSOpenPanel openPanel];
	[op setCanChooseDirectories:YES];
	[op setCanChooseFiles:NO];
	[op setAllowsMultipleSelection:NO];
	
	int ret = [op runModalForDirectory:nil file:nil types:nil];
	if(ret != NSOKButton) return;
	
	[o_prefGameDirectory setStringValue:[op filename]];
}

- (IBAction)removeIcon:(id)sender
{
	[o_prefGameIcon clearPath];
	[o_prefGameIcon setImage:[NSImage imageNamed:@"NSApplicationIcon"]];
}

- (IBAction)doubleClicked:(id)sender
{
	if([sender tag] == 0) {
		if([o_gameListTable numberOfSelectedRows] == 0) return;
		else [[mgr gameInfoAtIndex:[o_gameListTable selectedRow]] execute];
		[self updateTable];
	}
}

- (IBAction)beginCheckUpdate:(id)sender
{
	[self checkUpdate];
	
	if([[NSFileManager defaultManager] isWritableFileAtPath:[[NSBundle mainBundle] resourcePath]]
	   && [[NSFileManager defaultManager] isDeletableFileAtPath:[[NSBundle mainBundle] pathForResource:@"onscripter" ofType:nil inDirectory:nil]]) {
		if(isUpdateAvailable) {
			if(currentVersion < updateVersion) {
				if(![mgr terminateOK]) {
					NSBeginCriticalAlertSheet(LS(@"update onscripter"), LS(@"Cancel"), nil, nil, o_mainWindow, self, nil, nil, nil, LS(@"onscripter is busy"));
					return;
				}
				else {
					[o_updateMsg setStringValue:LS(@"update found")];
					char buf[12];
					sprintf(buf,"%d",updateVersion);
					buf[9] = buf[8];
					buf[8] = '-';
					buf[10] = 0;
					[o_updateVersion setStringValue:[NSString stringWithFormat:@"Ver. %s",buf]];
					[o_updateButton setEnabled:YES];
				}
			}
			else {
				NSBeginAlertSheet(LS(@"update onscripter"), LS(@"Cancel"), nil, nil, o_mainWindow, self, nil, nil, nil, LS(@"onscripter is up-to-date"));
				return;
			}
		}
		else {
			NSBeginCriticalAlertSheet(LS(@"update onscripter"), LS(@"Cancel"), nil, nil, o_mainWindow, self, nil, nil, nil, LS(@"update failure"));
			return;
		}
	} else {
		NSBeginCriticalAlertSheet(LS(@"update onscripter"), LS(@"Cancel"), nil, nil, o_mainWindow, self, nil, nil, nil, LS(@"no update permission"));
		return;
	}
	[o_updateButton setTitle:LS(@"update")];
	[o_updateCancelButton setEnabled:YES];
	[NSApp beginSheet:o_updatePane
	   modalForWindow:o_mainWindow
		modalDelegate:self
	   didEndSelector:@selector(sheetDidEnd:returnCode:contextInfo:)
		  contextInfo:@"Update"];
}

- (IBAction)beginUpdate:(id)sender
{
	if(updateFinished) {
		[NSApp endSheet:o_updatePane returnCode:0];
		[o_updatePane close];
		updateFinished = NO;
		return;
	}
	[o_updateMsg setStringValue:LS(@"downloading update")];
	[updateMgr beginDownload:updateURL];
	[o_updateVersion setStringValue:@""];
	[o_updateButton setEnabled:NO];
	updateFinished = NO;
}

- (IBAction)endUpdate:(id)sender
{
	[updateMgr cancelDownload];
	[NSApp endSheet:o_updatePane returnCode:0];
	[o_updatePane close];
	[o_updateProgress stopAnimation:self];
	[o_updateProgress setDoubleValue:0.0];
}

- (IBAction)showFontPanel:(id)sender
{
	NSFontPanel *fp = [NSFontPanel sharedFontPanel];
	[fp makeKeyAndOrderFront:nil];
}

- (void)addLogText:(NSString *)str
{
	if([str isEqualToString:lastLogStr]) {
		if(time(NULL)-lastLogTime < 1) return;
	}
	if(lastLogStr) [lastLogStr release];
	lastLogStr = [str retain];
	lastLogTime = time(NULL);
	/*[o_logView selectAll:nil];
	[o_logView setSelectedRange:NSMakeRange([o_logView selectedRange].length, 0)];
	[o_logView setEditable:YES];
	
	NSMutableAttributedString *attStr = [[NSMutableAttributedString alloc] initWithString:str];
	[attStr setAttributes:[NSDictionary dictionaryWithObject:[NSFont systemFontOfSize:10] forKey:NSFontAttributeName] range:NSMakeRange(0,[str length])];
	[o_logView insertText:attStr];
	if(bufferLine == 200) {
		NSTextStorage *st = [o_logView textStorage];
		NSRange range = [[st mutableString] rangeOfString:@"\n"];
		range.length = range.location+1;
		range.location = 0;
		[[st mutableString] replaceCharactersInRange:range withString:@""];
	}*/
	[logArr addObject:str];
	if([logArr count] == 200) [logArr removeObjectAtIndex:0];
	[o_logTableView reloadData];
	[o_logTableView scrollRowToVisible:[logArr count]-1];
	//[o_logView setEditable:NO];
	//[attStr release];
}

- (void)updateTable
{
	[o_gameListTable reloadData];
	[self tableViewSelectionDidChange:[NSNotification notificationWithName:nil object:o_gameListTable]];
}

- (void)savePrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	[pref setObject:[mgr saveGameList] forKey:@"GameList"];
	[pref setObject:[fontMgr saveFontList] forKey:@"FontList"];
	[pref setInteger:[[o_selectOnscripter selectedCell] tag] forKey:@"ONScripter"];
	[pref setObject:[o_executePath stringValue] forKey:@"ExecutePath"];
	[pref setInteger:[o_terminateOnExit state] forKey:@"TerminateOnExit"];
	[pref setInteger:[o_automaticUpdateCheck state] forKey:@"AutomaticUpdate"];
	[pref synchronize];
}

- (void)loadPrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	id obj;
	if(obj=[pref objectForKey:@"FontList"]) {
		[fontMgr loadFontList:obj];
	}
	if(obj=[pref objectForKey:@"GameList"]) {
		[mgr loadGameList:obj];
		[o_gameListTable reloadData];
	}
	if(obj=[pref objectForKey:@"ONScripter"]) {
		[o_selectOnscripter selectCellWithTag:[obj intValue]];
	}
	if(obj=[pref objectForKey:@"ExecutePath"]) {
		[o_executePath setStringValue:obj];
	}
	if(obj=[pref objectForKey:@"TerminateOnExit"]) {
		[o_terminateOnExit setState:[obj intValue]];
	}
	if(obj=[pref objectForKey:@"AutomaticUpdate"]) {
		[o_automaticUpdateCheck setState:[obj intValue]];
	}
}

- (id)fontManager
{
	return fontMgr;
}

- (NSString *)executePath
{
	if([[o_selectOnscripter selectedCell] tag] == 1) {
		if([[NSFileManager defaultManager] isExecutableFileAtPath:[o_executePath stringValue]])
			return [[o_executePath stringValue] stringByExpandingTildeInPath];
		else {
			[self addLogText:LS(@"onscripter not found")];
			return [[NSBundle mainBundle] pathForResource:@"onscripter" ofType:nil inDirectory:nil];
		}
	}
	else return [[NSBundle mainBundle] pathForResource:@"onscripter" ofType:nil inDirectory:nil];
}

- (void)checkUpdate
{
	NSString *url = [NSString stringWithString:@"http://tmkk.pv.land.to/onscripter/update.txt"];
	NSURLHandle *urlhandle = [NSURLHandle cachedHandleForURL:[NSURL URLWithString:url]];
	
	NSData *dat = [urlhandle resourceData];
	
	if([urlhandle status] != NSURLHandleLoadSucceeded) {
		isUpdateAvailable = NO;
	}
	else {
		if(updateURL) [updateURL release];
		updateVersion = [[[[[NSString alloc] initWithData:dat encoding:NSASCIIStringEncoding] autorelease] substringToIndex:9] intValue];
		updateURL = [[NSURL alloc] initWithString:[[[[[NSString alloc] initWithData:dat encoding:NSASCIIStringEncoding] autorelease] substringFromIndex:10] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]]];
		isUpdateAvailable = YES;
	}
}

- (void)setCurrentVersion
{
	FILE *fp = fopen([[[NSBundle mainBundle] pathForResource:@"version" ofType:@"txt" inDirectory:nil] UTF8String],"r");
	char buffer[10];
	fread(buffer,1,9,fp);
	fclose(fp);
	buffer[9] = 0;
	currentVersion = atoi(buffer);
}

- (void)updateCurrentVersion
{
	FILE *fp = fopen([[[NSBundle mainBundle] pathForResource:@"version" ofType:@"txt" inDirectory:nil] UTF8String],"w");
	char buffer[10];
	sprintf(buffer,"%d",updateVersion);
	fwrite(buffer,1,9,fp);
	fclose(fp);
	currentVersion = updateVersion;
}

- (void)setProgressLimit:(int)value
{
	[o_updateProgress setDoubleValue:0.0];
	[o_updateProgress setMaxValue:(double)value];
	[o_updateProgress startAnimation:self];
}

- (void)loadDataLength:(int)value
{
	[o_updateProgress setDoubleValue:[o_updateProgress doubleValue] + value];
}

- (void)updateFinished
{
	[o_updateProgress stopAnimation:self];
	[o_updateMsg setStringValue:LS(@"update complete")];
	[o_updateCancelButton setEnabled:NO];
	[self updateCurrentVersion];
	updateFinished = YES;
	[o_updateButton setTitle:LS(@"finished")];
	[o_updateButton setEnabled:YES];
}

- (void)updateFailed
{
	[o_updateMsg setStringValue:LS(@"download failure")];
	[o_updateButton setEnabled:YES];
}

- (void)sheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	ONSLGameInfo *gi;
	if([(NSString *)contextInfo isEqualToString:@"AddGame"]) {
		switch(returnCode) {
			case 0:
				gi = [[ONSLGameInfo alloc] initWithDelegate:self];
				[gi setTitle:[o_prefGameName stringValue]];
				[gi setPath:[o_prefGameDirectory stringValue]];
				NSFont *newFont = [[NSFontManager sharedFontManager] selectedFont];
				[gi setFont:newFont];
				[gi setImagePath:[o_prefGameIcon imgPath]];
				[gi setOtherArgs:[o_prefGameOption stringValue]];
				if([o_fullscreenMode state] == NSOnState) [gi setFullscreenMode:YES];
				else [gi setFullscreenMode:NO];
				if([o_varEditMode state] == NSOnState) [gi setVarEditMode:YES];
				else [gi setVarEditMode:NO];
				if([o_wheelAdvanceMode state] == NSOnState) [gi setWheelAdvanceMode:YES];
				else [gi setWheelAdvanceMode:NO];
				break;
			default:
				return;
		}
		[mgr addGameInfo:gi];
		[o_gameListTable reloadData];
		[gi release];
	}
	else if([(NSString *)contextInfo isEqualToString:@"EditGame"]) {
		switch(returnCode) {
			case 0:
				gi = [mgr gameInfoAtIndex:[o_gameListTable selectedRow]];
				[gi setTitle:[o_prefGameName stringValue]];
				[gi setPath:[o_prefGameDirectory stringValue]];
				NSFont *newFont = [[NSFontManager sharedFontManager] selectedFont];
				[gi setFont:newFont];
				[gi setImagePath:[o_prefGameIcon imgPath]];
				[gi setOtherArgs:[o_prefGameOption stringValue]];
				if([o_fullscreenMode state] == NSOnState) [gi setFullscreenMode:YES];
				else [gi setFullscreenMode:NO];
				if([o_varEditMode state] == NSOnState) [gi setVarEditMode:YES];
				else [gi setVarEditMode:NO];
				if([o_wheelAdvanceMode state] == NSOnState) [gi setWheelAdvanceMode:YES];
				else [gi setWheelAdvanceMode:NO];
				break;
			default:
				return;
		}
		[o_gameListTable reloadData];
	}
	else if([(NSString *)contextInfo isEqualToString:@"Terminate"]) {
		switch(returnCode) {
			case NSAlertDefaultReturn:
				gi = [mgr gameInfoAtIndex:[o_gameListTable selectedRow]];
				[gi terminate];
				break;
			default:
				return;
		}
	}
	else if([(NSString *)contextInfo isEqualToString:@"Quit"]) {
		switch(returnCode) {
			case NSAlertDefaultReturn:
				if([o_terminateOnExit state] == NSOnState) [mgr terminateAllGames];
				quitOK = YES;
				[NSApp terminate:nil];
				break;
			default:
				return;
		}
	}
	else if([(NSString *)contextInfo isEqualToString:@"RemoveGame"]) {
		switch(returnCode) {
			case NSAlertDefaultReturn:
				[mgr removeGameInfoAtIndex:[o_gameListTable selectedRow]];
				[o_gameListTable reloadData];
				break;
			default:
				return;
		}
	}
	else if([(NSString *)contextInfo isEqualToString:@"Update"]) {
	}
}

- (int)numberOfRowsInTableView:(NSTableView *)tableView
{
	if([tableView tag] == 0) return [mgr numberOfGames];
	else if([tableView tag] == 2) return [logArr count];
	return 0;
}

- (id)tableView:(NSTableView *)tableView
		objectValueForTableColumn:(NSTableColumn *)tableColumn
		row:(int)row
{
	if([tableView tag] == 0) {
		if([[tableColumn identifier] isEqual:@"Image"]) {
			NSImage *img;
			NSString *imgPath = [[mgr gameInfoAtIndex:row] imagePath];
			if([imgPath isEqualToString:@""]) img = [NSImage imageNamed:@"NSApplicationIcon"];
			else {
				if([[NSFileManager defaultManager] isReadableFileAtPath:imgPath]) {
					img = [[[NSImage alloc] initByReferencingFile:imgPath] autorelease];
					NSImageRep *rep = [img bestRepresentationForDevice:nil];
					float width = [rep pixelsWide];
					float height = [rep pixelsHigh];
					[img setSize:NSMakeSize(width,height)];
				}
				else img = [NSImage imageNamed:@"NSApplicationIcon"];
			}
			
			return img;
		}
		else if([[tableColumn identifier] isEqual:@"Info"]) {
			NSMutableParagraphStyle *style = [[[NSMutableParagraphStyle alloc] init] autorelease];
			[style setLineSpacing:5.0];
			BOOL selected = [tableView isRowSelected:row];
			NSMutableAttributedString *str = [[NSMutableAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[[mgr gameInfoAtIndex:row] title]]
																					attributes:[NSDictionary dictionaryWithObjects:
																								[NSArray arrayWithObjects:[NSFont boldSystemFontOfSize:13],style,selected?[NSColor selectedControlTextColor]:[NSColor blackColor], nil]
																												  forKeys:[NSArray arrayWithObjects:NSFontAttributeName,NSParagraphStyleAttributeName,NSForegroundColorAttributeName,nil]]];
			NSMutableAttributedString *path = [[NSMutableAttributedString alloc] initWithString:[[mgr gameInfoAtIndex:row] path]
																					 attributes:[NSDictionary dictionaryWithObjects:
																								 [NSArray arrayWithObjects:[NSFont systemFontOfSize:11],selected?[NSColor selectedControlTextColor]:[NSColor grayColor],style,nil]
																												   forKeys:[NSArray arrayWithObjects:NSFontAttributeName,NSForegroundColorAttributeName,NSParagraphStyleAttributeName,nil]]];
			[str appendAttributedString:[path autorelease]];
			return [str autorelease];
		}
		else if([[tableColumn identifier] isEqual:@"State"]) {
			if([[mgr gameInfoAtIndex:row] isPlaying]) return [NSImage imageNamed:@"playing2"];
			else return nil;
		}
		return nil;
	}
	else if([tableView tag] == 2) {
		return [logArr objectAtIndex:row];
	}
	return nil;
}

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	if([[aNotification object] tag] == 0) {
		if([o_gameListTable numberOfSelectedRows] != 0) {
			if([[mgr gameInfoAtIndex:[o_gameListTable selectedRow]] isPlaying]) {
				[[toolbarItems objectForKey:@"Launch"] setImage:[NSImage imageNamed:@"stop"]];
				[[toolbarItems objectForKey:@"Launch"] setLabel:LS(@"Stop")];
			}
			else {
				[[toolbarItems objectForKey:@"Launch"] setImage:[NSImage imageNamed:@"run"]];
				[[toolbarItems objectForKey:@"Launch"] setLabel:LS(@"Launch")];
			}
		}
	}
}

- (void)windowDidResize:(NSNotification *)aNotification
{

}

- (void)applicationDidFinishLaunching: (NSNotification *)notification
{
	[self loadPrefs];
	
	NSToolbarItem* item;
	item = [[NSToolbarItem alloc] initWithItemIdentifier:@"AddGame"];
	[item setLabel:LS(@"Add Game")];
	[item setPaletteLabel:LS(@"Add Game")];
	[item setImage:[NSImage imageNamed:@"add"]];
	[item setTarget:self];
	[item setAction:@selector(addGame:)];
	[toolbarItems setObject:item forKey:@"AddGame"];
	[item release];
	
	item = [[NSToolbarItem alloc] initWithItemIdentifier:@"EditGame"];
	[item setLabel:LS(@"Edit Game")];
	[item setPaletteLabel:LS(@"Edit Game")];
	[item setImage:[NSImage imageNamed:@"edit"]];
	[item setTarget:self];
	[item setAction:@selector(editGame:)];
	[toolbarItems setObject:item forKey:@"EditGame"];
	[item release];
	
	item = [[NSToolbarItem alloc] initWithItemIdentifier:@"DeleteGame"];
	[item setLabel:LS(@"Delete Game")];
	[item setPaletteLabel:LS(@"Delete Game")];
	[item setImage:[[NSWorkspace sharedWorkspace] iconForFileType:NSFileTypeForHFSTypeCode(kTrashIcon)]];
	[item setTarget:self];
	[item setAction:@selector(removeGame:)];
	[toolbarItems setObject:item forKey:@"DeleteGame"];
	[item release];
	
	item = [[NSToolbarItem alloc] initWithItemIdentifier:@"ShowLog"];
	[item setLabel:LS(@"Show Log")];
	[item setPaletteLabel:LS(@"Show Log")];
	[item setImage:[[NSWorkspace sharedWorkspace] iconForFile:@"/Applications/Utilities/Console.app"]];
	[item setTarget:o_logWindow];
	[item setAction:@selector(makeKeyAndOrderFront:)];
	[toolbarItems setObject:item forKey:@"ShowLog"];
	[item release];
	
	item = [[NSToolbarItem alloc] initWithItemIdentifier:@"Launch"];
	[item setLabel:LS(@"Launch")];
	[item setPaletteLabel:LS(@"Launch")];
	[item setImage:[NSImage imageNamed:@"run"]];
	[item setTarget:self];
	[item setAction:@selector(launchGame:)];
	[toolbarItems setObject:item forKey:@"Launch"];
	[item release];
	
	NSToolbar *theBar = [[NSToolbar alloc] initWithIdentifier:@"ONSLauncherToolbar"];
	[theBar setDelegate:self];
	[theBar setAllowsUserCustomization:YES];
	[theBar setAutosavesConfiguration:YES];
	[o_mainWindow setToolbar:theBar];
	[theBar release];
	
	[o_gameListTable registerForDraggedTypes:TableRowTypes];
	[o_gameListTable setDoubleAction:@selector(doubleClicked:)];
	[self setCurrentVersion];
	[o_mainWindow makeKeyAndOrderFront: self];
	NSGraphicsContext *gc = [NSGraphicsContext currentContext];
	[gc setImageInterpolation:NSImageInterpolationHigh];
	[o_gameListTable reloadData];
	if ([o_automaticUpdateCheck state] == NSOnState) {
		[self checkUpdate];
		if([[NSFileManager defaultManager] isWritableFileAtPath:[[NSBundle mainBundle] resourcePath]]
		   && [[NSFileManager defaultManager] isDeletableFileAtPath:[[NSBundle mainBundle] pathForResource:@"onscripter" ofType:nil inDirectory:nil]]) {
			if(isUpdateAvailable) {
				if(currentVersion < updateVersion) {
					if([mgr terminateOK]) {
						[o_updateMsg setStringValue:LS(@"update found")];
						char buf[12];
						sprintf(buf,"%d",updateVersion);
						buf[9] = buf[8];
						buf[8] = '-';
						buf[10] = 0;
						[o_updateVersion setStringValue:[NSString stringWithFormat:@"Ver. %s",buf]];
						[o_updateButton setEnabled:YES];
						[o_updateButton setTitle:LS(@"update")];
						[o_updateCancelButton setEnabled:YES];
						[NSApp beginSheet:o_updatePane
						   modalForWindow:o_mainWindow
							modalDelegate:self
						   didEndSelector:@selector(sheetDidEnd:returnCode:contextInfo:)
							  contextInfo:@"Update"];
					}
				}
			}
		}
	}
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	if(!quitOK && ![mgr terminateOK]) {
		NSBeginCriticalAlertSheet(LS(@"quit ONSLauncher"), @"OK", nil, @"Cancel", o_mainWindow, self, @selector(sheetDidEnd:returnCode:contextInfo:), nil, @"Quit", LS(@"quit ok?"));
		return NO;
	}
	
	[self savePrefs];
	return YES;
}

- (BOOL)windowShouldClose:(id)sender
{
	[NSApp terminate:nil];
	return NO;
}

- (void)changeFont:(id)sender
{
	if(![o_gameEditPane isVisible]) return;
	ONSLGameInfo *gi = [mgr gameInfoAtIndex:[o_gameListTable selectedRow]];
	NSFont *newFont = [[NSFontManager sharedFontManager] convertFont:[gi font]];
	[[NSFontManager sharedFontManager] setSelectedFont:newFont isMultiple:NO];
	[o_prefGameFontName setStringValue:[newFont displayName]];
	[o_prefGameFontName setFont:newFont];
}

- (unsigned int)validModesForFontPanel:(NSFontPanel *)fontPanel
{
	return NSFontPanelFaceModeMask|NSFontPanelCollectionModeMask;
}

- (NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)itemIdentifier willBeInsertedIntoToolbar:(BOOL)flag
{
	return [toolbarItems objectForKey:itemIdentifier];
}

-(NSArray *) toolbarDefaultItemIdentifiers:(NSToolbar*)toolbar
{
	return [NSArray arrayWithObjects:@"AddGame",@"EditGame",@"DeleteGame",NSToolbarFlexibleSpaceItemIdentifier,@"ShowLog",@"Launch",nil];
}

-(NSArray *) toolbarAllowedItemIdentifiers:(NSToolbar*)toolbar
{
	return [NSArray arrayWithObjects:@"AddGame",@"EditGame",@"DeleteGame",@"ShowLog",@"Launch",NSToolbarFlexibleSpaceItemIdentifier,NSToolbarSpaceItemIdentifier,NSToolbarSeparatorItemIdentifier,nil];
}

- (BOOL)validateToolbarItem:(NSToolbarItem *)theItem
{
	if([o_gameListTable numberOfSelectedRows] == 0) {
		if([[theItem itemIdentifier] isEqualToString:@"EditGame"]) return NO;
		if([[theItem itemIdentifier] isEqualToString:@"DeleteGame"]) return NO;
		if([[theItem itemIdentifier] isEqualToString:@"Launch"]) return NO;
	}
	return YES;
}

-(BOOL)tableView:(NSTableView *)tv writeRows:(NSArray*)rows toPasteboard:(NSPasteboard*)pboard
{
	[pboard declareTypes:TableRowTypes owner:self];
	[pboard setPropertyList:rows forType:TableRowType];
	return YES;
}

-(NSDragOperation)tableView:(NSTableView*)tv validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)op
{
	NSPasteboard *pboard=[info draggingPasteboard];
	
	if (op == NSTableViewDropAbove && [pboard availableTypeFromArray:TableRowTypes] != nil) {
		return NSDragOperationMove;
	} else {
		return NSDragOperationNone;
	}
}

-(BOOL)tableView:(NSTableView*)tv acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)op
{
	NSPasteboard *pboard=[info draggingPasteboard];
	NSEnumerator *e=[[pboard propertyListForType:TableRowType] objectEnumerator];
	NSNumber *number;
	NSMutableArray *upperArray=[NSMutableArray arrayWithArray:[[mgr gameInfoArray] subarrayWithRange:NSMakeRange(0,row)]];
	NSMutableArray *lowerArray=[NSMutableArray arrayWithArray:[[mgr gameInfoArray] subarrayWithRange:NSMakeRange(row,([[mgr gameInfoArray] count] - row))]];
	NSMutableArray *middleArray=[NSMutableArray arrayWithCapacity:0];
	id object;
	int i;
	
	if (op == NSTableViewDropAbove && [pboard availableTypeFromArray:TableRowTypes] != nil) {
		
		while ((number=[e nextObject]) != nil) {
			object=[[mgr gameInfoArray] objectAtIndex:[number intValue]];
			[middleArray addObject:object];
			[upperArray removeObject:object];
			[lowerArray removeObject:object];
		}
		
		[[mgr gameInfoArray] removeAllObjects];
		
		[[mgr gameInfoArray] addObjectsFromArray:upperArray];
		[[mgr gameInfoArray] addObjectsFromArray:middleArray];
		[[mgr gameInfoArray] addObjectsFromArray:lowerArray];
		
		[o_gameListTable reloadData];
		[o_gameListTable deselectAll:nil];
		
		for (i=[upperArray count];i<([upperArray count] + [middleArray count]);i++) {
			[o_gameListTable selectRowIndexes:[NSIndexSet indexSetWithIndex:i] byExtendingSelection:[o_gameListTable allowsMultipleSelection]];
		}
		return YES;
	} else {
		return NO;
	}
}

@end
